'----------------------------------------------------------------------------
'
' This program is provided as Freeware for any and all uses under the
' modified BSD license. 
'
' Copyright 2008, Elba Corp.
' All rights reserved.
' http://www.zbasic.net
'
' Redistribution and use in source and binary forms, with or without 
' modification, are permitted provided that all of the following conditions
' are met:
'
' * Redistributions of source code must retain the above copyright notice,
'   this list of conditions and the following disclaimer.
'
'   and/or other materials provided with the distribution.
' * Neither the name of Elba Corp. nor the names of its contributors 
'   may be used to endorse or promote products derived from this software 
'   without specific prior written permission.
'
' THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
' IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
' ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
' LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
' CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
' SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
' INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
' CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
' ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
' POSSIBILITY OF SUCH DAMAGE.
'
'----------------------------------------------------------------------------
'
' This module performs reads from and writes to an SPI EEPROM such
' as the Atmel AT25256A or the STMicro M95512.  As written,
' the code does not support devices that cannot perform byte
' writes (such as the AT25HP512) but it could easily be extended
' to do so.

' This code assumes that that no protection bits are set in the EEPROM.

'----------------------------------------------------------------------------
' EEPROM command codes - compatible with the AT25256A, et al
Const WRSR_CMD as Byte = 1
Const WRITE_CMD as Byte = 2
Const READ_CMD as Byte = 3
Const WRDI_CMD as Byte = 4
Const RDSR_CMD as Byte = 5
Const WREN_CMD as Byte = 6

' This page size matches the AT25256A.  For the STMicro M95512, use 128.
' for other devices, consult the datasheet to determine the page size.
Const PAGE_SIZE as UnsignedInteger = 64

'----------------------------------------------------------------------------
'
'' InitSPI_EEPROM
'
' Initialize the SPI channel for EEPROM operations.
'
Sub InitSPI_EEPROM(ByVal chan as Byte, ByVal pin as Byte)
	Call OpenSPI(chan, 0, pin)
End Sub

'----------------------------------------------------------------------------
'
'' ReadSPI_EEPROM
'
' Read one or more bytes of data from the EEPROM beginning at a specified address.
' The data read is written to RAM beginning at the value given by the 'addrRAM'
' parameter.
'
Sub ReadSPI_EEPROM(ByVal chan as Byte, ByVal addrEEPROM as UnsignedInteger, _
		ByVal addrRAM as UnsignedInteger, ByVal count as UnsignedInteger)
	Dim wrBuf(1 to 3) as Byte
	Dim rdBuf() as Byte Based addrRAM

	If (count > 0) Then
		wrBuf(1) = READ_CMD
		wrBuf(2) = HiByte(addrEEPROM)
		wrBuf(3) = LoByte(addrEEPROM)
		Call SPICmd(chan, 3, wrBuf, count, rdBuf)
	End If
End Sub

'----------------------------------------------------------------------------
'
'' WriteSPI_EEPROM
'
' Write one or more bytes of data to the EEPROM beginning at a specified address.
' The data to write is read from RAM beginning at the value given by the 'addrRAM'
' parameter.
'
Sub WriteSPI_EEPROM(ByVal chan as Byte, ByVal addrEEPROM as UnsignedInteger, _
		ByVal addrRAM as UnsignedInteger, ByVal count as UnsignedInteger)
	Dim wrBuf(1 to PAGE_SIZE + 3) as Byte
	Dim rdBuf() as Byte Based addrRAM

	Do While (count > 0)
		Dim partial as UnsignedInteger

		' send a write enable command
		wrBuf(1) = WREN_CMD
		Call SPICmd(chan, 1, wrBuf, 0, 0)

		' compute how many bytes can be written on this pass
		partial = PAGE_SIZE - (addrEEPROM Mod PAGE_SIZE)
		If (partial > count) Then
			partial = count
		End If

		' prepare the data to be written
		wrBuf(1) = WRITE_CMD
		wrBuf(2) = HiByte(addrEEPROM)
		wrBuf(3) = LoByte(addrEEPROM)
		Call MemCopy(wrBuf.DataAddress + 3, addrRAM, partial)

		' write the data
		Call SPICmd(chan, 3 + partial, wrBuf, 0, 0)

		' prepare for the next pass
		addrEEPROM = addrEEPROM + partial
		addrRAM = addrRAM + partial
		count = count - partial

		' wait for the device to become ready
		Call waitReady(chan)
	Loop
End Sub

'----------------------------------------------------------------------------
'
'' waitReady
'
' This routine waits in a loop polling the EEPROM unitl it
' indicates that it is no longer busy.
'
Private Sub waitReady(ByVal chan as Byte)
	Dim cmd as Byte
	Dim stat as Byte
	
	' wait for the device to become ready
	Do
		cmd = RDSR_CMD
		Call SPICmd(chan, 1, cmd, 1, stat)
	Loop While (CBool(stat And &H01))
'	Debug.Print "status=0x";CStrHex(stat)
End Sub

' uncomment this line to include the test driver in the compiled code
'#define INCLUDE_TEST_DRIVER

#ifdef INCLUDE_TEST_DRIVER
'----------------------------------------------------------------------------
'
' This is a simple test driver for the SPI EEPROM routines.
'
Sub Main()
	' define the starting data value
	Const dataStart as Byte = &H55
	Const dataOfst as Byte = &H01

	' define the SPI channel
	Const chan as Byte = 2

	' define the starting address for the test
	Const addr as UnsignedInteger = 5

#if Option.TargetCode = "Native"
	Const pin as Byte = B.4
#else
	Const pin as Byte = 5
#endif

	' initialize the SPI channel
	Call InitSPI_EEPROM(chan, pin)

	' write some data to the EEPROM
	Dim i as Integer
	Dim buf(1 to 256) as Byte
	Dim data as Byte
	data = dataStart
	For i = LoBound(buf) to HiBound(buf)
		buf(i) = data
		data = data + dataOfst
	Next i
'	Debug.Print "Writing"
	Call WriteSPI_EEPROM(chan, addr, buf.DataAddress, CUInt(SizeOf(buf)))

	' "erase" the data buffer
	Call MemSet(buf.DataAddress, SizeOf(buf), &H55)

	' read the data back
'	Debug.Print "Reading"
	Call ReadSPI_EEPROM(chan, addr, buf.DataAddress, CUInt(SizeOf(buf)))

	' confirm that the data is correct
	Dim OK as Boolean
	OK = True
	data = dataStart
	For i = LoBound(buf) to HiBound(buf)
		If (buf(i + 1) <> data) Then
			Debug.Print "Error at address 0x"; CStrHex(i); " val=0x"; CStrHex(buf(i)); " s/b "; CStrHex(data)
			OK = False
			Exit For
		End If
		data = data + dataOfst
	Next i
	If (OK) Then
		Debug.Print "EEPROM test completed successfully"
	End If
End Sub
#endif
